%{
/* Wolf - modified to read constraints. */

#include "ipp.h"

/* default yywrap function - always treat EOF as an EOF  */
#define opswrap() 1 

int bracket_count = 0;

%}

a [Aa]
b [Bb]
c [Cc]
d [Dd]
e [Ee]
f [Ff]
g [Gg]
h [Hh]
i [Ii]
j [Jj]
k [Kk]
l [Ll]
m [Mm]
n [Nn]
o [Oo]
p [Pp]
q [Qq]
r [Rr]
s [Ss]
t [Tt]
u [Uu]
v [Vv]
w [Ww]
x [Xx]
y [Yy]
z [Zz]
DIGIT [[:digit:]]
POINT \.
NUMBER-NORM {DIGIT}+({POINT}{DIGIT}+)?
NUMBER-BARE {POINT}{DIGIT}+
NUMBER      ({NUMBER-BARE}|{NUMBER-NORM})
OP [+-*/]
COMP (">="|"<="|"=")
VAR \?[a-zA-Z][a-zA-Z0-9\-_]* 
NAME :?[a-zA-Z][a-zA-Z0-9\-_]* 
CONSTRAINT_VAR {VAR}\'?(\[({VAR}|{NAME})(,({VAR}|{NAME}))*\])?

%x COMMENT OVERREAD CONSTRAINT_SECTION

%%

"("[ \t]*{t}{e}{s}{t} {
  bracket_count = 1;
  BEGIN CONSTRAINT_SECTION;
  return(TEST_TOK);
}

"("[ \t]*{s}{e}{t}  {
  bracket_count = 1;
  BEGIN CONSTRAINT_SECTION;
  return(SET_TOK);
}

"("[ \t]*{i}{n}{f}{l}{u}{e}{n}{c}{e}  {
  bracket_count = 1;
  BEGIN CONSTRAINT_SECTION;
  return(INFLUENCE_TOK);
}

"("  {  return('('); }

")"  {  return(')'); }

"-"  {  return('-'); }

"("[ \t]*{i}{n}"-"{p}{a}{c}{k}{a}{g}{e}  {  bracket_count = 1;
 BEGIN OVERREAD; }

{d}{e}{f}{i}{n}{e}  {  return(DEFINE_TOK); }

{d}{o}{m}{a}{i}{n}  {  return(DOMAIN_TOK); }

"("[ \t]*":"{e}{x}{t}{e}{n}{d}{s}  {  bracket_count = 1;
 BEGIN OVERREAD; }

":"{r}{e}{q}{u}{i}{r}{e}{m}{e}{n}{t}{s}  {  return(REQUIREMENTS_TOK); }

":"{t}{y}{p}{e}{s}  {  return(TYPES_TOK); }

":"{c}{o}{n}{s}{t}{a}{n}{t}{s}  {  return(CONSTANTS_TOK); }

":"{d}{o}{m}{a}{i}{n}"-"{v}{a}{r}{i}{a}{b}{l}{e}{s}  {  
  bracket_count = 1;
  BEGIN CONSTRAINT_SECTION;
  return(DOMAIN_VARIABLES_TOK); 
} 

":"{p}{r}{e}{d}{i}{c}{a}{t}{e}{s}  { return(PREDICATES_TOK); }

"("[ \t]*":"{t}{i}{m}{e}{l}{e}{s}{s} {  bracket_count = 1;
 BEGIN OVERREAD; }

":"{a}{c}{t}{i}{o}{n}  {  return(ACTION_TOK); }

":"{a}{x}{i}{o}{m}  {  bracket_count = 1; BEGIN OVERREAD; }

":"{p}{a}{r}{a}{m}{e}{t}{e}{r}{s}  {  return(PARAMETERS_TOK); }

":"{v}{a}{r}{s}  { return(VARS_TOK); }

":"{p}{r}{e}{c}{o}{n}{d}{i}{t}{i}{o}{n}  {  return(PRECONDITION_TOK); }

":"{c}{o}{n}{t}{e}{x}{t}  { return(PRECONDITION_TOK); }

":"{e}{f}{f}{e}{c}{t}  {  return(EFFECT_TOK); }

":"{i}{m}{p}{l}{i}{e}{s}  {  return(EFFECT_TOK); }

":"{f}{l}{u}{e}{n}{t}"-"{p}{a}{r}{a}{m}{e}{t}{e}{r}{s} { return(FLUENT_PARAMETERS_TOK); }

{a}{n}{d}  {  return(AND_TOK); }

{n}{o}{t}  {  return(NOT_TOK); }

{w}{h}{e}{n}  {  return(WHEN_TOK); }

{i}{m}{p}{l}{y} {  return(IMPLY_TOK); }

{o}{r} {  return(OR_TOK); }

{f}{o}{r}{a}{l}{l} {  return(FORALL_TOK); }

{e}{x}{i}{s}{t}{s} {  return(EXISTS_TOK); }

<CONSTRAINT_SECTION>{NUMBER} {
  strcpy(yylval.string, yytext);
  return(NUMBER_TOK);
}

<CONSTRAINT_SECTION>{CONSTRAINT_VAR} {
  strcpy(yylval.string, strupcase(yytext));
  return(VARIABLE);
}

<CONSTRAINT_SECTION>"*" {
  strcpy(yylval.string, yytext);
  return('*');
}

<CONSTRAINT_SECTION>"/" {
  strcpy(yylval.string, yytext);
  return('/');
}

<CONSTRAINT_SECTION>"+" {
  strcpy(yylval.string, yytext);
  return('+');
}
<CONSTRAINT_SECTION>"-" {
  strcpy(yylval.string, yytext);
  return('-');
}

<CONSTRAINT_SECTION>{COMP} {
  strcpy(yylval.string, yytext);
  return(COMP_TOK);
}

<CONSTRAINT_SECTION>"(" {
  bracket_count++;
  return('(');
}

<CONSTRAINT_SECTION>")" {
  bracket_count--;
  if (bracket_count == 0)
    BEGIN INITIAL;
  return(')');
}

"="  {  return(EQUAL_TOK); }

<CONSTRAINT_SECTION>"-"[ \t]*{f}{l}{u}{e}{n}{t} { 
  return(FLUENT_TOK); 
}

<CONSTRAINT_SECTION>"-"[ \t]*{f}{l}{o}{a}{t} { 
  return(FLOAT_TOK); 
}

:?[a-zA-Z][a-zA-Z0-9\-_]* {  
strcpy(yylval.string, strupcase( yytext ) ); return(NAME); }

\?[a-zA-Z][a-zA-Z0-9\-_\[\]]* {  
strcpy(yylval.string, strupcase( yytext ) ); return(VARIABLE); }

"-"[ \t]*"("[ \t]*{e}{i}{t}{h}{e}{r} { return(EITHER_TOK); }

"-"[ \t]*{f}{l}{u}{e}{n}{t} { return(FLUENT_TOK); }

"-"[ \t]*{f}{l}{o}{a}{t} { return(FLOAT_TOK); }

"-"[ \t]*[a-zA-Z][a-zA-Z0-9\-_\[\]]*  {   
 strcpy(yylval.string, rmdash(strupcase(yytext))); return(TYPE); }

\;[^\n]*\n  {  lineno++; } 
\;[^\n]*  {  /* this will hold only in files that end with
		   a comment but no linefeed */ } 

<CONSTRAINT_SECTION>\;[^\n]*\n  {  lineno++; } 

<OVERREAD>\;[^\n]*\n  {  lineno++; } 

<COMMENT>(.^\")*\n    {  lineno++; }  ;

<INITIAL>\" { BEGIN COMMENT;}

<COMMENT>\" { BEGIN INITIAL;}

\n    {  lineno++; } 

<OVERREAD>(.^\(\))*\n  {  lineno++; }

<OVERREAD>[^\(\)]  {  }

<OVERREAD>\(  {  BEGIN OVERREAD; bracket_count++; }

<OVERREAD>\)  {  BEGIN OVERREAD; bracket_count--; 
  if (!bracket_count) BEGIN INITIAL; }

([1-9][0-9]*)|"0" {
  strcpy(yylval.string, yytext);
  return(NUMBER_TOK);
}

. {}

<CONSTRAINT_SECTION>. {}

<CONSTRAINT_SECTION>\n { lineno++; }

%%


