%{
/* Wolf - modified to handle constraints. */

#ifdef YYDEBUG
  extern int yydebug=1;
#endif


#include <stdio.h>
#include <string.h> 
#include "ipp.h"

#ifndef SCAN_ERR
#define SCAN_ERR
#define DOMDEF_EXPECTED            0
#define DOMAIN_EXPECTED            1
#define DOMNAME_EXPECTED           2
#define LBRACKET_EXPECTED          3
#define RBRACKET_EXPECTED          4
#define DOMDEFS_EXPECTED           5
#define REQUIREM_EXPECTED          6
#define TYPEDLIST_EXPECTED         7
#define LITERAL_EXPECTED           8
#define PRECONDDEF_UNCORRECT       9
#define TYPEDEF_EXPECTED          10
#define CONSTLIST_EXPECTED        11
#define PREDDEF_EXPECTED          12 
#define NAME_EXPECTED             13
#define VARIABLE_EXPECTED         14
#define ACTIONFUNCTOR_EXPECTED    15
#define ATOM_FORMULA_EXPECTED     16
#define EFFECT_DEF_EXPECTED       17
#define NEG_FORMULA_EXPECTED      18
#define NOT_SUPPORTED             19
#define ACTION                    20
  
#endif

  char *adl_string[] = { "nothing", "literal", "and", "or","imply", 
		  "not", "exists", 
		  "forall", "(", ")", 
		  "variable", "endnot",
		  NULL };     

static char *errmsg[] = {
  "domain definition expected",
  "'domain' expected",
  "domain name expected",
  "'(' expected",
  "')' expected",
  "additional domain definitions expected",
  "requirements (e.g. ':STRIPS') expected",
  "typed list of <%s> expected",
  "literal expected",
  "uncorrect precondition definition",
  "type definition expected",
  "list of constants expected",
  "predicate definition expected",
  "<name> expected",
  "<variable> expected",
  "action functor expected",
  "atomic formula expected",
  "effect definition expected",
  "negated atomic formula expected",
  "requirement %s not supported by this IPP version",  
  "action definition is not correct",
  NULL
};


void opserr( int errno, char *par );
int yyerror( char * msg );
int yylex();

int supported( char *str );

int act_err;
char *act_err_par = NULL;
op_list cur_op = NULL;
fact_list cur_fl;

%}

%start file

/* This may seem strange, but enables us to use the types known from
   ipp.h */
%union {
  char string[256];
  token tok;
  fact_list fl;
  op_list ol;
  token_list tl;
}


%type <fl> adl_effect
%type <fl> adl_effect_star
%type <fl> adl_goal_description
%type <fl> adl_goal_description_star
%type <tl> literal_term
/*%type <tl> literal_name*/
/*%type <fl> literal_name_plus*/
%type <tl> term_star
%type <fl> typed_list_name
%type <fl> typed_list_variable
%type <tok> term
%type <tl> atomic_formula_term
/*%type <tl> atomic_formula_name*/

%type <tl> constraint
%type <tl> cons_expr

%type <tl> name_plus
%type <tl> variable_star
%type <tok> predicate

%type <fl> pred_def_star
%type <tl> pred_def
%type <fl> typed_list_domain_var_decl
%type <tl> domain_var_decl
%type <tok> float_type

%type <tl> cons_expr

%token DEFINE_TOK
%token DOMAIN_TOK
%token REQUIREMENTS_TOK
%token DOMAIN_VARIABLES_TOK
%token PREDICATES_TOK
%token TYPES_TOK
%token EITHER_TOK
%token FLUENT_TOK /* These tokens are for */
%token FLOAT_TOK  /* constrained vars     */
%token <string> COMP_TOK
%token TEST_TOK
%token SET_TOK
%token INFLUENCE_TOK
%token <string> NUMBER_TOK
%token CONSTANTS_TOK
%token ACTION_TOK
%token AXIOM_TOK
%token VARS_TOK
%token PRECONDITION_TOK
%token FLUENT_PARAMETERS_TOK
%token PARAMETERS_TOK
%token EFFECT_TOK
%token AND_TOK
%token NOT_TOK
%token WHEN_TOK
%token FORALL_TOK
%token IMPLY_TOK
%token OR_TOK
%token EXISTS_TOK
%token EQUAL_TOK
%token <string> NAME
%token <string> VARIABLE
%token <string> TYPE

%left '-' '+'
%left '/' '*'
%left NEG

%%

/* numbers in comments are only useful for the programmer
   and reader of the yacc source code. Each terminal or
   non-terminal symbol and (which is not nice) each
   program section in {brackets} has a number and may
   be reached by the $ operator - the numbers only
   help prevent counting all the time and producing
   ugly fat spots on the monitor */

/**********************************************************************/
file:
{ /*opserr( DOMDEF_EXPECTED, NULL );*/ }
domain_definition 
;
/* can be extended to support 'addenda' and similar stuff */

/**********************************************************************/
domain_definition : 
'(' DEFINE_TOK  
domain_name       
{ 
  /* initialize typetree */
  global_type_tree_list = new_type_tree_list( STANDARD_TYPE );
}
optional_domain_defs 
{
  ;
}
;

/**********************************************************************/
domain_name :
'(' DOMAIN_TOK  NAME  ')' 
{ 
  gdomain_name = new_token( strlen($3)+1 );
  strcpy( gdomain_name, $3);
}
;

/**********************************************************************/
optional_domain_defs:
')'  /* end of domain */
|
require_def optional_domain_defs
|
constants_def optional_domain_defs
|
types_def optional_domain_defs
|
action_def optional_domain_defs
|
domain_variables_def optional_domain_defs
|
predicates_def optional_domain_defs
;


/**********************************************************************/
require_def:
'(' REQUIREMENTS_TOK 
{ /*opserr( REQUIREM_EXPECTED, NULL );*/ }
NAME
{ 
  if ( !supported( $4 ) )
    {
      /*opserr( NOT_SUPPORTED, $4 );*/
      yyerror(NULL);
    }
}
require_key_star
')'
;

require_key_star:
/* empty */
|
NAME
{ 
  if ( !supported( $1 ) )
    {
      /*opserr( NOT_SUPPORTED, $1 );*/
      yyerror(NULL);
    }
}
require_key_star
;

/**********************************************************************/
types_def:
'(' TYPES_TOK
{ /*opserr( TYPEDEF_EXPECTED, NULL );*/ }
typed_list_name 
')'
{ 
  add_to_type_tree( $4, main_type_tree() ); 
}
; 

/**********************************************************************/
constants_def:
'(' CONSTANTS_TOK
{ /*opserr( CONSTLIST_EXPECTED, NULL );*/ }
typed_list_name
')'
{ 
  orig_constant_list = $4;
}
;


/**********************************************************************
 * actions and their optional definitions
 **********************************************************************/

action_def:
'(' ACTION_TOK 
{ /*opserr( ACTION, NULL );*/ }
NAME
{ 
  cur_op = new_op_list( $4 );
}
optional_action_sections
{
  cur_op->next = loaded_ops;
  loaded_ops = cur_op; 
}
;

/**********************************************************************/

optional_action_sections:
')' /* end of action sections */
|
param_def optional_action_sections
|
fluent_param_def optional_action_sections
|
action_def_body optional_action_sections
;

param_def:
PARAMETERS_TOK
{ cur_op->params = NULL; }
|
PARAMETERS_TOK '(' typed_list_variable ')'
{
  fact_list f;
  cur_op->params = $3;
  for( f=cur_op->params; f; f = f->next )
    cur_op->number_of_real_params++; /* to be able to distinguish
					params from :VARS */
}

/**********************************************************************/

fluent_param_def:
FLUENT_PARAMETERS_TOK
{ cur_op->fluent_params = NULL; }
|
FLUENT_PARAMETERS_TOK '(' variable_star ')'
{
  cur_op->fluent_params = $3;
}

/**********************************************************************/

action_def_body:
/* empty */
|
VARS_TOK '(' typed_list_variable ')' action_def_body
{
  fact_list f;
  /* add vars as parameters */
  if ( cur_op->params )
    {
      for( f = cur_op->params; f->next; f = f->next )
	;
      f->next = $3;
      f = f->next;
    }
  else
    /* Wolf - this used to be f = $3 which didn't look right to me */
    cur_op->params = $3;
}
|
PRECONDITION_TOK
{
  /* no effect */
}
|
PRECONDITION_TOK
adl_goal_description
{ 
  if (cur_op->hooks == NULL) {
    cur_op->hooks = $2;
  }
  else {
    fact_list f;
    for (f = cur_op->hooks; f->next; f = f->next) {
	 ;
    }
    f->next = $2;
  }
}
action_def_body
|
EFFECT_TOK
{
  /* no effect */
}
|
EFFECT_TOK
adl_effect
{ 
  if (cur_op->hooks == NULL) {
    cur_op->hooks = $2;
  }
  else {
    fact_list f;
    for (f = cur_op->hooks; f->next; f = f->next) {
	 ;
    }
    f->next = $2;
  }
}
action_def_body
;

/**********************************************************************
 * predicates defn
/**********************************************************************/

/* stores predicates in a fact_list. Each predicate starts with a token list
 * of length one (the pred symbol itself). That's how to separate them! All
 * args have a type. */

predicates_def:
'(' PREDICATES_TOK
{ ; /* placeholder */ }
pred_def_star
')'
{ predicates_list = $4 }
;

pred_def_star:
/* empty */
{ $$ = NULL; }
|
pred_def pred_def_star
{ 
  $$ = new_fact_list();
  $$->item = $1;
  $$->next = $2;
}
;

pred_def:
'(' NAME typed_list_variable ')'
{
  token_list tl;
  fact_list fl;

  $$ = new_token_list();
  $$->item = new_token(strlen($2) + 1);
  strcpy($$->item, $2);

  tl = $$;

  for (fl = $3; fl; fl = fl->next) {
    tl->next = fl->item->next;
    fl->item->next = NULL;
    tl = tl->next;
  }

  free_complete_fact_list(fl);
}
;

/**********************************************************************
 * timeless defn: things that are always true
/**********************************************************************/
/*
timeless_def:
'(' TIMELESS_TOK
literal_name_plus ')'
{
  timeless_list = $3;
}
*/
/**********************************************************************
 * domain variable defn
/**********************************************************************/

domain_variables_def:
'(' DOMAIN_VARIABLES_TOK
typed_list_domain_var_decl
')'
{ 
  fact_list fl;

  if ($3) {
    for (fl = $3; fl->next; fl = fl->next) {
	 ;
    }
    fl->next = domain_vars_list;
    domain_vars_list = $3; 
  }
}
|
'(' DOMAIN_VARIABLES_TOK ')'
{ }
;

/* I'm not handling any type but fluent and float in this section! */
/* For now I'm not handling preset values */
typed_list_domain_var_decl:
domain_var_decl float_type
{
  token_list opt_num;

  $$ = new_fact_list();
  $$->item = $1;
  opt_num = $1->next;

  $$->item->next = new_token_list();
  $$->item->next->item = $2;

  $$->item->next->next = opt_num;

  $$->next = NULL;
}
|
domain_var_decl float_type typed_list_domain_var_decl   
/* end of list for one type */
{
  /* store the optional number in the domain variable declaration */
  token_list opt_num;

  $$ = new_fact_list();
  $$->item = $1;
  opt_num = $1->next;

  $$->item->next = new_token_list();
  $$->item->next->item = $2;

  $$->item->next->next = opt_num;

  $$->next = $3;
}
|
domain_var_decl typed_list_domain_var_decl    
/* a list element (gets type from next one) */
{
  token_list opt_num;
  token s = $2->item->next->item;
  int len = strlen( s ) + 1;

  $$ = new_fact_list();
  $$->item = $1;
  opt_num = $1->next;

  $$->item->next = new_token_list();
  $$->item->next->item = new_token(len);
  strcpy($$->item->next->item, s);

  $$->item->next->next = opt_num;

  $$->next = $2;
}
;

domain_var_decl:
VARIABLE
{
  $$ = new_token_list();
  $$->item = new_token(strlen($1) + 1);
  strcpy($$->item, $1);
}
|
'(' VARIABLE NUMBER_TOK ')'
{
  $$ = new_token_list();
  $$->item = new_token(strlen($2) + 1);
  strcpy($$->item, $2);

  $$->next = new_token_list();
  $$->next->item = new_token(strlen($3) + 1);
  strcpy($$->next->item, $3);
}
;

float_type:
FLOAT_TOK
{ 
  $$ = new_token(strlen(FLOAT_STR) + 1); 
  strcpy($$, FLOAT_STR);
}
|
FLUENT_TOK
{ 
  $$ = new_token(strlen(FLUENT_STR) + 1); 
  strcpy($$, FLUENT_STR);
}
;

/**********************************************************************
 * goal description providing full ADL
/**********************************************************************/

adl_goal_description: /* this provides only an easy-to-handle
			 structure for pl1-expressions which will
			 be preprocessed later. returns fact_list */
literal_term
{ 
  /* Wolf - the better to merge preconds and effects, I'm adding an initial
	precond/effect flag! */
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token(strlen(PRECOND_STR) + 1);
  strcpy($$->item->item, PRECOND_STR);
  $$->item->next = $1;
}
|
'(' AND_TOK
adl_goal_description_star
')'
{ 
  fact_list f;
  
  $$ = make_adl_fact( AND_CONST );
  $$->next = $3; 
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
|
'(' OR_TOK
adl_goal_description_star
')'
{ 
  fact_list f;
  
  $$ = make_adl_fact( OR_CONST );
  $$->next = $3; 
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
|
'(' NOT_TOK
adl_goal_description
')'
{ 
  fact_list f;
  
  $$ = make_adl_fact( NOT_CONST );
  $$->next = $3; 
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( ENDNOT_CONST );
}
|
'(' IMPLY_TOK
adl_goal_description
adl_goal_description
')'
{ 
  fact_list f;
  
  $$ = make_adl_fact( OR_CONST );
  $$->next = make_adl_fact( NOT_CONST );
  $$->next->next = $3; 
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( ENDNOT_CONST);
  f->next->next = $4;
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
|
'(' EXISTS_TOK
'('
typed_list_variable
')'
adl_goal_description
')'
{ 
  fact_list f;
  
  $$ = f = make_adl_fact( EXISTS_CONST );
  f->next = make_adl_fact( LBRACK_CONST );
  f->next->next = $4; 
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
  f = f->next;
  f->next = $6;
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
|
'(' FORALL_TOK
'('
typed_list_variable
')'
adl_goal_description
')'
{ 
  fact_list f;
  
  $$ = f = make_adl_fact( FORALL_CONST );
  f->next = make_adl_fact( LBRACK_CONST );
  f->next->next = $4; 
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
  f = f->next;
  f->next = $6;
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
;


adl_goal_description_star:
{
  $$ = NULL;
}
|
adl_goal_description adl_goal_description_star
{
  fact_list f;

  $$ = $1;
  if ( !$$ )
    $$ = $2;
  else
    {
      for ( f = $$; f->next; f = f->next )
	;
      f->next = $2;
    }
}


/**********************************************************************
 * effects as allowed in pddl are saved in IPP data structures
 *********************************************************************/

adl_effect: /* describes everything after the keyword :effect */
literal_term
{
  /* Wolf - the better to merge preconds and effects, I'm adding an initial
	precond/effect flag! */
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token(strlen(EFFECT_STR) + 1);
  strcpy($$->item->item, EFFECT_STR);
  $$->item->next = $1;
}
|
'(' AND_TOK
adl_effect_star /* return effect_list */
')'
{
  fact_list f;
  
  $$ = make_adl_fact( AND_CONST );
  $$->next = $3;
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}  
|
'(' NOT_TOK
literal_term
')'
{ 
  $$ = make_adl_fact( NOT_CONST );
  /* Wolf - the better to merge preconds and effects, I'm adding an initial
	precond/effect flag! */
  $$->next = new_fact_list();
  $$->next->item = new_token_list();
  $$->next->item->item = new_token(strlen(EFFECT_STR) + 1);
  strcpy($$->next->item->item, EFFECT_STR);
  $$->next->item->next = $3;
  $$->next->next = make_adl_fact( ENDNOT_CONST );
}
|
'(' FORALL_TOK
'('
typed_list_variable
')'
adl_effect
')'
{
  fact_list f;
  
  $$ = f = make_adl_fact( FORALL_CONST );
  f->next = make_adl_fact( LBRACK_CONST );
  f->next->next = $4; 
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
  f = f->next;
  f->next = $6;
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
|
'(' WHEN_TOK
adl_goal_description
adl_effect
')'
{
  fact_list f;
  
  $$ = make_adl_fact( OR_CONST );
  $$->next = make_adl_fact( NOT_CONST );
  $$->next->next = $3; 
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( ENDNOT_CONST);
  f->next->next = $4;
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
;

adl_effect_star:
/* empty */
{ 
  $$ = NULL; 
}
|
adl_effect
adl_effect_star
{
  fact_list f;

  $$ = $1;
  if ( !$$ )
    $$ = $2;
  else
    {
      for ( f = $$; f->next; f = f->next )
	   ;
      f->next = $2;
    }
}
;



/**********************************************************************
 * some expressions used in many different rules
 **********************************************************************/
/*
literal_name_plus:
literal_name
{
  $$ = new_fact_list();
  $$->item = $1;
}
|
literal_name literal_name_plus
{
  $$ = new_fact_list();
  $$->item = $1;
  $$->next = $2;
}
;

literal_name:
atomic_formula_name
{
  $$ = $1;
}
|
'(' NOT_TOK atomic_formula_name ')'
{
  token t = new_token(strlen($2->item) + 1 + 1);
  sprintf(t, "%s%s", "!", $2->item);
  my_free($2->item);
  $2->item = t;
  $$ = $2;
}
;

atomic_formula_name:
'(' name_plus ')'
{
  $$ = $2;
}
*/
/**************************************************************************/

/* Note that I modified this slightly (Wolf). Don't use it without checking
   that you're getting the right meaning!! */
literal_term:
atomic_formula_term
{
  $$ = $1;
}
;

/**********************************************************************/

atomic_formula_term:
'('
predicate 
term_star
')'
{ 
  $$ = new_token_list();
  $$->item = new_token( strlen($2)+1 );
  strcpy( $$->item, $2 );
  $$->next = $3;
}
|
constraint
{
  $$ = $1;
}
;

/**********************************************************************/

term_star:
/* empty */
{ $$ = NULL; }
|
term
term_star
{
  $$ = new_token_list();
  $$->item = new_token( strlen($1)+1 );
  strcpy( $$->item, $1 );
  $$->next = $2;
}
;

/**********************************************************************/

term:
NAME
{ 
  $$ = new_token( strlen($1)+1 );
  strcpy( $$, $1 );
}
|
VARIABLE
{ 
  $$ = new_token( strlen($1)+1 );
  strcpy( $$, $1 );
}
;

/**********************************************************************/

variable_star:
/* empty */
{ $$ = NULL; }
|
VARIABLE variable_star
{
  $$ = new_token_list();
  $$->item = new_token(strlen($1) + 1);
  strcpy($$->item, $1);
  $$->next = $2;
}

/**********************************************************************/

name_plus:
NAME
{
  $$ = new_token_list();
  $$->item = new_token( strlen($1)+1 );
  strcpy( $$->item, $1 );
}
|
NAME name_plus
{
  $$ = new_token_list();
  $$->item = new_token( strlen($1)+1 );
  strcpy( $$->item, $1 );
  $$->next = $2;
}


/**********************************************************************/

typed_list_name:     /* returns fact_list */
/* empty */
{ $$ = NULL; }
|
NAME EITHER_TOK name_plus ')' typed_list_name
{ /* this is a very, very special case... it may mean that
     a. a parameter has one of the types in name_plus
     b. a type is subtype of one of the following types in name_plus 
     => for both possibilities we use the same solution:
     build a new type name: either_name1_name2_..._namen and check
     if a type with this name already exists. If so then NAME has this type,
     if not then build a new type, include it in the type tree and
     return NAME with the new type 
     The new artificial type is not a subtype of OBJECT because its own
     elements must already be instances of a subtype of OBJECT */
  token s;
  token_list t;
  type_tree tt;
  type_tree_list rootl, *st;
  
  s = new_token( MAX_LENGTH );
  strcpy( s, EITHER_STR );
  for ( t = $3; t; t = t->next )
    {
      strcat( s, CONNECTOR );
      strcat( s, t->item );
    }
  tt = NULL;
  for ( rootl = global_type_tree_list; rootl; rootl = rootl->next )
    if ( (tt = find_branch( s, rootl->item )) )
      break;
  if ( !tt )
    { /* the type doesn't exist yet, so build it */
      rootl = new_type_tree_list( s );
      rootl->next = global_type_tree_list;
      st = &(rootl->item->sub_types);
      for ( t = $3; t; t = t->next )
	{
	  if ( (tt = find_branch( t->item, main_type_tree() )) )
	    {
	      *st = new_type_tree_list( NULL );
	      (*st)->item = tt;
	      st = &((*st)->next);
	    }
	}  
    }

  /* now do the simple stuff: return a name with a (quite complicated)
     type */
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  $$->item->next->item = s;
  $$->next = $5;
}
|
NAME TYPE typed_list_name   /* end of list for one type */
{
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  $$->item->next->item = new_token( strlen($2)+1 );
  strcpy( $$->item->next->item, $2 );
  $$->next = $3;
}
|
NAME typed_list_name        /* a list element (gets type from next one) */
{
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  if ( $2 )   /* another element (already typed) is following */
    {
      char* s = $2->item->next->item;
      int   l = strlen( s ) + 1;
      $$->item->next->item = new_token( l );
      strcpy( $$->item->next->item, s ); /* same type as the next one */
      $$->next = $2;
    }
  else /* no further element - it must be an untyped list */
    {
      $$->item->next->item = new_token( strlen(STANDARD_TYPE)+1 );
      strcpy( $$->item->next->item, STANDARD_TYPE );
      $$->next = $2;
    }
}
;

/***********************************************/

typed_list_variable:     /* returns fact_list */
/* empty */
{ $$ = NULL; }
|
VARIABLE EITHER_TOK name_plus ')' typed_list_variable
{ /* this is a very, very special case... it may mean that
     a parameter has one of the types in name_plus
     => build a new type name: either_name1_name2_..._namen and check
     if a type with this name already exists. If so then NAME has this type,
     if not then build a new type, include it in the type tree and
     return NAME with the new type 
     The new artificial type is not a subtype of OBJECT because its own
     elements must already be instances of a subtype of OBJECT */
  token s;
  token_list t;
  type_tree tt;
  type_tree_list rootl, *st;
  
  s = new_token( MAX_LENGTH );
  strcpy( s, EITHER_STR );
  for ( t = $3; t; t = t->next )
    {
      strcat( s, CONNECTOR );
      strcat( s, t->item );
    }
  tt = NULL;
  for ( rootl = global_type_tree_list; rootl; rootl = rootl->next )
    if ( (tt = find_branch( s, rootl->item )) )
      break;
  if ( !tt )
    { /* the type doesn't exist yet, so build it */
      rootl = new_type_tree_list( s );
      rootl->next = global_type_tree_list;
      st = &(rootl->item->sub_types);
      for ( t = $3; t; t = t->next )
	{
	  if ( (tt = find_branch( t->item, main_type_tree() )) )
	    {
	      *st = new_type_tree_list( NULL );
	      (*st)->item = tt;
	      st = &((*st)->next);
	    }
	}  
      global_type_tree_list = rootl;
    }

  /* now do the simple stuff: return a name with a (quite complicated)
     type */
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  $$->item->next->item = s;
  $$->next = $5;
}
|
VARIABLE TYPE typed_list_variable   /* end of list for one type */
{
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  $$->item->next->item = new_token( strlen($2)+1 );
  strcpy( $$->item->next->item, $2 );
  $$->next = $3;
}
|
VARIABLE typed_list_variable    /* a list element (gets type from next one) */
{
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  if ( $2 )   /* another element (already typed) is following */
    {
      char* s = $2->item->next->item;
      int   l = strlen( s );
      $$->item->next->item = new_token( l+1 );
      strcpy( $$->item->next->item, s ); /* same type as the next one */
      $$->next = $2;
    }
  else /* no further element - it must be an untyped list */
    {
      $$->item->next->item = new_token( strlen(STANDARD_TYPE)+1 );
      strcpy( $$->item->next->item, STANDARD_TYPE );
      $$->next = $2;
    }
}
;

/**********************************************************************/

predicate:
NAME
{ 
  $$ = new_token( strlen($1)+1 );
  strcpy( $$, $1 );
}
|
EQUAL_TOK
{ 
  $$ = new_token( strlen(EQ_STR)+1 );
  strcpy( $$, EQ_STR );
}
;

/**********************************************************/
/*  CONSTRAINT STUFF                                      */
/**********************************************************/

constraint:
TEST_TOK
cons_expr COMP_TOK cons_expr ')'
{
  token_list tl;

  $$ = new_token_list();
  $$->item = new_token(strlen(TEST_STR) + 1);
  strcpy($$->item, TEST_STR);
  $$->next = $2;

  for (tl = $$->next; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen($3) + 1);
  strcpy(tl->next->item, $3);
  tl->next->next = $4;
}
|
SET_TOK
VARIABLE cons_expr ')'
{
  $$ = new_token_list();
  $$->item = new_token(strlen(SET_STR) + 1);
  strcpy($$->item, SET_STR);

  $$->next = new_token_list();
  $$->next->item = new_token(strlen($2) + 1);
  strcpy($$->next->item, $2);

  $$->next->next = new_token_list();
  $$->next->next->item = new_token(strlen("=") + 1);
  strcpy($$->next->next->item, "=");
  $$->next->next->next = $3;
}
|
INFLUENCE_TOK
VARIABLE cons_expr ')'
{
  token_list tl;

  $$ = tl = new_token_list();
  tl->item = new_token(strlen(INFLUENCE_STR) + 1);
  strcpy(tl->item, INFLUENCE_STR);

  tl->next = new_token_list();
  tl = tl->next;
  tl->item = new_token(strlen($2) + 1);
  strcpy(tl->item, $2);

  tl->next = new_token_list();
  tl = tl->next;
  tl->item = new_token(strlen("=") + 1);
  strcpy(tl->item, "=");

  tl->next = new_token_list();
  tl = tl->next;
  tl->item = new_token(strlen($2) + 1);
  strcpy(tl->item, $2);

  tl->next = new_token_list();
  tl = tl->next;
  tl->item = new_token(strlen("+") + 1);
  strcpy(tl->item, "+");

  tl->next = $3;
}
;

cons_expr:
NUMBER_TOK
{
  $$ = new_token_list();
  $$->item = new_token(strlen($1) + 1);
  strcpy($$->item, $1);
}
|
VARIABLE
{
  $$ = new_token_list();
  $$->item = new_token(strlen($1) + 1);
  strcpy($$->item, $1);
}
|
cons_expr '*' cons_expr
{
  token_list tl;

  $$ = $1;
  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen("*") + 1);
  strcpy(tl->next->item, "*");
  tl->next->next = $3;
}
|
cons_expr '/' cons_expr
{
  token_list tl;

  $$ = $1;
  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen("/") + 1);
  strcpy(tl->next->item, "/");
  tl->next->next = $3;
}
|
cons_expr '+' cons_expr
{
  token_list tl;

  $$ = $1;
  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen("+") + 1);
  strcpy(tl->next->item, "+");
  tl->next->next = $3;
}
|
cons_expr '-' cons_expr
{
  token_list tl;

  $$ = $1;
  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen("-") + 1);
  strcpy(tl->next->item, "-");
  tl->next->next = $3;
}
|
'-' cons_expr %prec NEG
{
  $$ = new_token_list();
  $$->item = new_token(strlen("-") + 1);
  strcpy($$->item, "-");
  $$->next = $2;
}
|
'(' cons_expr ')'
{
  token_list tl = new_token_list();
  tl->item = new_token(strlen("(") + 1);
  strcpy(tl->item, "(");
  tl->next = $2;
  $$ = tl;

  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen(")") + 1);
  strcpy(tl->next->item, ")");
}
;


%%

#include "lex.ops.c"

/**********************************************************************
 * Functions
 **********************************************************************/

/* 
call	bison -pops -bscan-ops scan-ops.y
*/
void opserr( int errno, char *par )
{
  act_err = errno;
  if ( act_err_par )
    my_free( act_err_par );
  if ( par )
    {
      act_err_par = new_token( strlen(par)+1 );
      strcpy( act_err_par, par);
    }
  else
    act_err_par = NULL;
}
  
int yyerror( char *msg )
{
  fflush( stdout );
  fprintf(stderr, "\n%s: syntax error in line %d, '%s':\n", act_filename, 
		lineno, yytext );
  if ( act_err_par )
    {
      fprintf(stderr, errmsg[act_err], act_err_par );
      fprintf(stderr, "\n" );
    }
  else
    fprintf(stderr,"%s\n", errmsg[act_err] );

  OUTPUT_FILE;
  exit( 1 );
}


void load_ops_file( char *filename )
{
  FILE *fp;/* pointer to input files */

  /* open operator file */
  if( ( fp = fopen( filename, "r" ) ) == NULL ) {
    fprintf(stderr, "\nipp: can't find operator file: %s\n\n", filename );
    OUTPUT_FILE;
    exit( 1 );
  }
  act_filename = filename;
  lineno = 1; 
  yyin = fp;
  yyparse();
  fclose( fp );/* and close file again */
}


void print_ops( op_list op )
{ 
  
  op_list    i_op;

  for ( i_op = op; i_op; i_op = i_op->next ) {
    
    printf( "\n%s \n", i_op->name );

    printf( "v: (%d real params) ", i_op->number_of_real_params );
    print_factlist( i_op->params );

    printf( "\nc: ");
    print_tokenlist( i_op->fluent_params );

    printf( "\np: ");
    print_clauselist( i_op->hooks );
  }
}


/***************************************************************/
/* those ones should be in somewhere in the IPP kernel         */

int supported( char *str )
{
  char *sup[] = { ":STRIPS", ":NEGATION", ":EQUALITY",":TYPING", 
		  ":CONDITIONAL-EFFECTS", ":DISJUNCTIVE-PRECONDITIONS", 
		  ":EXISTENTIAL-PRECONDITIONS", ":UNIVERSAL-PRECONDITIONS", 
		  ":QUANTIFIED-PRECONDITIONS", ":ADL",
		  ":DOMAIN-AXIOMS", ":SUBGOAL-THROUGH-AXIOMS",
		  NULL };     
  int i;
  
  for(i=0; sup[i]; i++ )
    if ( strcmp( sup[i], str ) == SAME )
      return TRUE;
  return FALSE;
}
